unit gfglTimeEstimator;
{copyright Greg Lorriman 2002 greg@lorriman.com http://www.lorriman.com

Free for use by any one.

DISCLAIMER :

The Author of this Software expressly disclaims any warranty for the SOFTWARE
PRODUCT. The SOFTWARE PRODUCT and any related documentation is
provided "as is" without warranty of any kind, either express or implied, including,

without limitation, the implied warranties or merchantability, fitness for a particular
purpose, or noninfringement. The entire risk arising out of use or performance of the SOFTWARE PRODUCT remains with you.

NO LIABILITY FOR DAMAGES.
In no event shall the author of this Software be liable for any damages whatsoever
(including, without limitation, damages for loss of business profits, business
interruption, loss of business information, or any other pecuniary loss) arising out

of the use of or inability to use this product, even if the Author of this Software
has been advised of the possibility of such damages. Because some states /
jurisdictions do not allow the exclusion or limitation of liability for consequential
or incidental damages, the above limitation may not apply to you. In this case the
Author of this Software will only be liable for the amount of money you spent for this SOFTWARE PRODUCT.

}

interface

{$IFNDEF VER110}
{$IFNDEF VER100}
{$IFNDEF VER93}
{$IFNDEF VER90}
{$IFNDEF VER80}
{$IFNDEF VER70}
{$DEFINE D4UP}
{$ENDIF}
{$ENDIF}
{$ENDIF}
{$ENDIF}
{$ENDIF}
{$ENDIF}

uses
   Windows, Messages, SysUtils, Classes;

type
   ETimeEstimator = class(Exception);
   TAdjustment = (aNone, aDynamic, aAdd, aSubtract);
{
   Sampling weighted time estimator for intergral and continuous ranges.
   "Continuous" means that the Position values do not have to be whole
   numbers.

   Modelled on TProgressBar so you can copy TProgressBar code; except
   that you must call the Start method to begin estimating. Samples are
   taken when Position is modified or StepBy, Stepit or Update
   are called. Typically the component would be used with a timer to
   avoid a "jerky" estimate display if Adjustment = aSubtract.

   Set Max, Position and Step to initial values. Min is unused. Step defaults
   to 10 as with TProgressBar, which may not be desired. Call Start to begin
   estimating. Read TimeLeft to get estimate. TimeLeft cannot be formatted
   by usual DateTime format functions as the date portion is a count of
   days; for your convenience TimeLeftAsString is supplied that formats the
   value correctly. Set Adjustment to compensate for lapses in time between updates
   to Position.

   Use Elapsed (or ElapsedAsString) and Throughput to complete the progress display.
   String properties are programmed to give a result suited to UI display before,
   during and after the end of processing.

      Change SampleSize if the domain warrants a different sampling size.
   All Properties can be modified safely after estimation has begun.
   TimeLeft and TimeLeftAsString are present in the object inspector for
   the purpose of visibility; changing their values has no effect.

   Setting Position to a value beyond Max is undefined. Decrementing Position causes an exception.
   }
   TgfglTimeEstimator = class(TComponent)
   private
    { Private declarations }
      FSamples: TList;
      FStartTime: TDateTime;
      FLastTime: TDateTime;
      FLastPosition: double;
      FPosition: double;
      FMax: double;
      FMin: double;
      FStep: double;
      FSampleSize: integer;
      FAdjustment: TAdjustment;
      FVersion: string;
      FHelp: TStrings;
    FIgnoreRegression: boolean;

      procedure SetMax(const Value: double);
      procedure SetPosition(const Value: double);
      procedure SetMin(const Value: double);
      procedure SetStep(const Value: double);
      procedure setTimeLeft(const Value: TDateTime);
      procedure clearMarks;
      procedure setTimeLeftAsString(const Value: string);
      procedure SetSampleSize(const Value: integer);
      procedure SetVersion(const Value: string);
      procedure SetHelp(const Value: TStrings);
      procedure SetThroughput(const Value: double);
      function getElapsed: TDateTime;
      procedure setElapsed(const Value: TDateTime);
      procedure setElapsedAsString(const Value: string);
      function getOtherStuff: string;
      procedure setOtherStuff(const Value: string);
    procedure SetIgnoreRegression(const Value: boolean);

   protected
    { Protected declarations }
      function getTimeLeft: TDateTime; virtual;
      function getElapsedAsString: string; virtual;
      function getTimeLeftAsString: string; virtual;
      function getThroughput: double; virtual;
      function formatLength(len: TDateTime): string; virtual;
      function getComponentHomePage: string; virtual;

   public
    { Public declarations }

      constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;

      {Resets estimator. Call this to begin estimation.}
      procedure Start;
      {Sets Position (and triggers sampling) and returns time estimate

      For Delphi 4 and above : overloaded method also gives
      throughput. This method avoids two calculations of estimate which would
      happen if both TimeLeft and Throughput were individually read, which may be important in
      performance situations especially if the sample size were large.
      For most situations the performance effect, which is minimal, is very unlikely to warrant use of this method.

      }
      function Update(position: double): TDateTime; {$IFDEF D4UP} overload; {$ENDIF}
{$IFDEF D4UP}
      function update(position: double; var throughput: double): TDateTime; overload;
{$ENDIF}
      {Increments Position (and triggers sampling) by Step}
      procedure Stepit;
      {Increments Position (and triggers sampling) by value}
      procedure Stepby(value: double);

   published
      //ignore this property - must be published
      property ComponentHomePage: string read getComponentHomePage;
      {
      If a reading from TimeLeft is taken when there has been no update to Position then the
      value can be adjusted by adding or substracting the time elapsed since the last update.
      Defaults to aDynamic.

      aNone = No adjustment

      aAdd = useful for compensating for progress that may significantly stall

      aSubtract = makes TimeLeft progressive; it will keep counting down

      aDynamic = if lapse in update to position is twice the average then add, else subtract.

      aSubtract tends to a smooth progression of TimeLeft towards zero. aAdd makes the TimeLeft value jump somewhat but corrects for significant stalls
      in the update of Position which may be useful in some circumstances. aDynamic has the advantages
      of both.
      }
      property Adjustment: TAdjustment read FAdjustment write FAdjustment default aDynamic;
     {Set to target value.}
      property Max: double read FMax write SetMax;
     {Unused, place holder for TProgressBar compatability}
      property Min: double read FMin write SetMin;
      {The value by which Position is increased when calling Step method. Cannot be less than 0.}
      property Step: double read FStep write SetStep;
      {Set this to your starting point. Assigning to this property after calling Start
      causes a sample to be recorded. Decrementing raises an exception.}
      property Position: double read FPosition write SetPosition;
      {The sample size. Defaults to 30, which is adequate for most purposes. Cannot be less than 0.

      Time estimation will occur even if their are less samples than this value, so don;t change
      this setting other than to reflect a domain behaviour.}
      property SampleSize: integer read FSampleSize write SetSampleSize default 30;
      {Returns time estimate. -1 if no time estimate available (no samples). Returns 0 when Position = Max or greater.

      The value is not a date; the date portion of the value is a count of days so you cannot use standard DateToString
      format functions. Use TimeLeftAsString instead, or look at the source to compose your own. }
      property TimeLeft: TDateTime read getTimeLeft write setTimeLeft stored false;
      {Throughput in units of Position, per a second.}
      property Throughput: double read getThroughput write SetThroughput stored false;
      {Formats TimeLeft taking into consideration that the date portion is not a date but "days".

      Formatting is - "[Days ####] time" - ("days" only appears if necessary)

      Also see FormatDateTimeLength which is supplied as a stand alone function.}
      property TimeLeftAsString: string read getTimeLeftAsString write setTimeLeftAsString stored false;
      {Easy enough to calculate without a component; present for your convenience.}
      property Elapsed: TDateTime read getElapsed write setElapsed stored false;
      {
       Setting Position to a value lower than its current value is undefined and therefore
       an exception is raised. To allow for regression set IgnoreRegression to true; setting
       Position to a lower value will be ignored without an exception. IgnoreRegression defaults to true.
      }
      property IgnoreRegression : boolean read FIgnoreRegression write SetIgnoreRegression default true;
      {Gives a string formatted to reflect that the date portion of Elapsed is a
      measure of time rather than a date.

      Same formatting as TimeLeftAsString.}
      property ElapsedAsString: string read getElapsedAsString write setElapsedAsString stored false;
      property Version: string read FVersion write SetVersion stored false;
      property Help: TStrings read FHelp write SetHelp stored false;
      property OtherStuff: string read getOtherStuff write setOtherStuff stored false;
   end;

{formats TDateTime as as length of time, rather than from a date, resulting in
"[Days ####] time" - ("days" only appears if necessary)}
function FormatDateTimeLength(len: TDateTime): string;

procedure Register;

implementation

type
   TMark = class
   private
      FTimeDif: TDateTime;
      FPosDif: double;
   public
      constructor create(posDif: double; timeDif: TDateTime);
   end;

{ TgfglSimplePredictor }

procedure TgfglTimeEstimator.clearMarks;
var
   i: integer;
begin
   for i := 0 to FSamples.count - 1 do
      Tobject(FSamples[i]).free;
   FSamples.clear;
end;

constructor TgfglTimeEstimator.create(AOwner: TComponent);
const helpStr = 'Sampling weighted time estimator for intergral and continuous ranges. '#13#10+
'"Continuous" means that the Position values do not have to be whole '#13#10+
'numbers.'#13#10+
''#13#10+
'Modelled on TProgressBar so you can copy TProgressBar code; except '#13#10+
'that you must call the Start method to begin estimating. Samples are '#13#10+
'taken when Position is modified or StepBy, Stepit or Update are '#13#10+
'called. Typically the component would be used with a timer to avoid a '#13#10+
'"jerky" estimate display if Adjustment = aSubtract.'#13#10+
''#13#10+
'Set Max, Position and Step to initial values. Min is unused. Step '#13#10+
'defaults to 10 as with TProgressBar, which may not be desired. Call '#13#10+
'Start to begin estimating. Read TimeLeft to get estimate. TimeLeft '#13#10+
'cannot be formatted by usual DateTime format functions as the date '#13#10+
'portion is a count of days; for your convenience TimeLeftAsString is '#13#10+
'supplied that formats the value correctly. Use Adjustment to '#13#10+
'compensate for lapses in time between updates to Position.'#13#10+
''#13#10+
'Use Elapsed (or ElapsedAsString) and Throughput to complete the '#13#10+
'progress display. String properties are programmed to give a result '#13#10+
'suited to UI display before, during and after the end of processing.'#13#10+
''#13#10+
'Change SampleSize if the domain warrants a different sampling size. '#13#10+
'All Properties can be modified safely after estimation has begun. '#13#10+
'TimeLeft and TimeLeftAsString are present in the object inspector for '#13#10+
'the purpose of visibility; changing their values has no effect.'#13#10+
''#13#10+
'Setting position to a value beyond Max is undefined. Decrementing '#13#10+
'Position is undefined.'#13#10;

begin
   inherited;
   FSamples := TList.create;
   FMax := 100;
   FStep := 10;
   FSampleSize := 30;
   FVersion := '.2 beta for D1/2/3/4/5/6';
   FHelp := TStringlist.create;
   FHelp.text := helpstr;
   FAdjustment := aDynamic;
   FIgnoreRegression:=true;
end;

destructor TgfglTimeEstimator.destroy;
begin
   clearMarks;
   FSamples.free;
   FHelp.free;
   inherited;
end;

function TgfglTimeEstimator.formatLength(len: TDateTime): string;
begin
   result := FormatDateTimeLength(len);
end;

function TgfglTimeEstimator.getComponentHomePage: string;
begin
   result := 'http://www.lorriman.com/programming/components.htm';
end;

const
   SecsPerDay = 24 * 60 * 60;

function calcThroughput(timeLeft: TDatetime; RangeRemaining: double): double;
begin
   result := 0;
   if (RangeRemaining > 0) and (TimeLeft > 0) then
      result := RangeRemaining / (timeLeft * secsPerDay);
end;

function TgfglTimeEstimator.getThroughput: double;
begin
   result := calcThroughput(timeLeft, FMax - FPosition);
end;

function TgfglTimeEstimator.getElapsed: TDateTime;
begin
   result := now - FStartTime;
end;

function TgfglTimeEstimator.getElapsedAsString: string;
begin
   result := '-';
   if FStartTime = 0 then EXIT;
   result := FormatLength(getElapsed);
end;

function TgfglTimeEstimator.getOtherStuff: string;
begin
   result:='http://www.lorriman.com';
end;

function TgfglTimeEstimator.getTimeLeft: TDateTime;
var
   i: integer;
   factor: integer;
   meanDivider: integer;
   totalledAvgTime: TDateTime;
   avgTime: TDatetime;
   offset: TDateTime;
   posdif: double;
   timedif: TDateTime;
   adjust : TAdjustment;
begin
   result := -1;

   meanDivider := 0;
   totalledAvgTime := 0;

   if FPosition>=FMax then begin result:=0; EXIT; end;

   for i := 0 to FSamples.count - 1 do begin
      factor := FSampleSize + 1 - i;
      inc(meanDivider, factor);
      timedif := TMark(FSamples[i]).FTimeDif;
      posdif := TMark(FSamples[i]).FPosDif;
      totalledAvgTime := totalledAvgTime + timedif / posdif * factor;
   end;

   if (totalledAvgTime * meandivider) = 0 then EXIT;

   avgTime := totalledAvgTime / meanDivider;
   result := (FMax - FLastPosition) * avgTime;

   if FAdjustment <> aNone then begin
      offset := now - FLastTime;
      adjust:=FAdjustment;
      if FAdjustment=aDynamic then begin
         if offset>(2*avgTime) then
            adjust:=aAdd
         else
            adjust:=aSubtract;
      end;
      case adjust of
         aAdd :  result := result + offset;
         aSubtract : result := result - offset;
      end;
   end;
end;

function FormatDateTimeLength(len: TDateTime): string;
var
   days: integer;
begin
   result := '';
   days := trunc(len);
   if days <> 0 then result := result + 'Days ' + inttostr(days) + ' ';
   result := result + timetostr(len);
end;

function TgfglTimeEstimator.getTimeLeftAsString: string;
var
   timeLeft: TDatetime;
begin
   result := '-';
   timeLeft := gettimeLeft;
   if timeleft = -1 then EXIT;
   result := FormatDateTimeLength(timeleft);
end;

procedure TgfglTimeEstimator.setElapsed(const Value: TDateTime);
begin

end;

procedure TgfglTimeEstimator.setElapsedAsString(const Value: string);
begin

end;

procedure TgfglTimeEstimator.SetHelp(const Value: TStrings);
begin

end;

procedure TgfglTimeEstimator.SetMax(const Value: double);
begin
   FMax := Value;
end;

procedure TgfglTimeEstimator.SetMin(const Value: double);
begin
   FMin := Value;
end;

procedure TgfglTimeEstimator.setOtherStuff(const Value: string);
begin

end;

procedure TgfglTimeEstimator.SetPosition(const Value: double);
var
   timenow: TDateTime;
begin
   if value<FPosition then begin
      if FIgnoreRegression then EXIT;
      raise ETimeEstimator.create('Cannot set Position to less than itself');
   end;

   if (Fposition <> value) then begin
      FPosition := Value;
      timeNow := now;
      FSamples.insert(0, TMark.create(value - FLastPosition, timeNow - FLastTime));
      while FSamples.count > FSampleSize do begin
         Tobject(FSamples[FSamples.count - 1]).free;
         FSamples.delete(FSamples.count - 1);
      end;
      FLastTime := timeNow;
      FLastPosition := value;
   end;
end;

procedure TgfglTimeEstimator.SetSampleSize(const Value: integer);
begin
   if Value >= 0 then
      FSampleSize := Value;
end;

procedure TgfglTimeEstimator.SetStep(const Value: double);
begin
   if value>=0 then
      FStep := Value;
end;

procedure TgfglTimeEstimator.SetThroughput(const Value: double);
begin

end;

procedure TgfglTimeEstimator.setTimeLeft(const Value: TDateTime);
begin

end;

procedure TgfglTimeEstimator.setTimeLeftAsString(const Value: string);
begin

end;

procedure TgfglTimeEstimator.SetVersion(const Value: string);
begin

end;

procedure TgfglTimeEstimator.start;
begin
   ClearMarks;
   FStartTime := now;
   FSamples.clear;
   FLastTime := FStartTime;
   FLastPosition := FPosition;
end;

procedure TgfglTimeEstimator.stepby(value: double);
begin
   position := position + value;
end;

procedure TgfglTimeEstimator.stepit;
begin
   position := position + FStep;
end;

function TgfglTimeEstimator.update(position: double): TDateTime;
begin
   self.position := position;
   result := getTimeLeft;
end;

{ TMark }

constructor TMark.create(posDif: double; timeDif: TDateTime);
begin
   FTimeDif := timedif;
   FPosDif := posdif;
end;

function TgfglTimeEstimator.Update(position: double;
   var throughput: double): TDateTime;
begin
   setPosition(position);
   result := getTimeLeft;
   throughput := calcThroughput(result, FMax - FPosition);
end;

procedure Register;
begin
   RegisterComponents('Custom', [TgfglTimeEstimator]);
end;

procedure TgfglTimeEstimator.SetIgnoreRegression(const Value: boolean);
begin
  FIgnoreRegression := Value;
end;

end.

